import sys
import os
import platform
platformName=platform.system()
pandaVersion='1.9.4'

arch=platform.architecture()[0]

if arch=='64bit':
    arch='-x64'
else:
    arch=''

PROFILE=False

import shutil
from distutils.core import setup

from Cython.Distutils import build_ext, Extension
from Cython.Compiler import Options
from Cython.Build import cythonize

Options.annotate=True

import subprocess
import itertools

from time import time

# for parallelism
useOpenMP=False

releaseBuild=True

import platform
platformName=platform.system()
win=platformName=='Windows'

import numpy

verbose=False

test=False

def arg(name):
    if name in sys.argv:
        sys.argv.remove(name)
        return True
    return False

def remove(path):
    if os.path.exists(path):
        if verbose: print(("removing "+path))
        os.remove(path)


def protoBuild():
    protoDirs=['ncore','api','plugin']
    for root, dirs, files in itertools.chain.from_iterable(os.walk(d) for d in protoDirs):
        for f in files:
            name,ext=os.path.splitext(f)
            if ext=='.proto':
                lang="python"
                cmd="protoc -I="+root+" --"+lang+"_out="+root+" "+os.path.join(root,f)
                print(cmd)
                err=subprocess.call(cmd, shell=True)
                if err: return err
    return False

def buildExtensions():
    sys.argv.append('build_ext')
    sys.argv.append('--inplace')
    #sys.argv.append('--build-lib')
    c=os.curdir
    args={}

    # configure OS dependant compiler settings
    # you may need to adjust the paths to match your systems
    libs=['libpanda']
    
    args['extra_compile_args']=[]
    args['extra_link_args']=[]
    
    pyd_appends=''
    if win:
        import cython
        s=cython.__file__.split('\\')
        sitePath='\\'.join(s[:-1])
        cythonPath=sitePath+'\\cython.py'
        cythonBin=['python',cythonPath]#'cython.pyc'

        pandaPath='C:\\Panda3D-' + pandaVersion + arch + '\\'
        cythonBin=['python',cythonPath]#'cython.pyc'
        ext=".pyd"

        major, minor = sys.version_info.major, sys.version_info.minor
        if (major, minor) == (3, 7):
            pyd_appends='.cp'+str(major)+str(minor)

        if arch=='-x64':
            pyd_appends+='-win_amd64'
        
        args['include_dirs']=[pandaPath+'include']
        args['include_dirs'].extend([numpy.get_include()])
        #args['include_dirs'].extend([r"plugin\printer\loaders\aop"])
        args['library_dirs']=[pandaPath+r'python\libs', pandaPath+'lib']
        args['extra_compile_args']=['/EHsc']
        args['libraries']=libs
    elif mac: # Mac
        cythonBin=['cython']
        ext=".so"
        args['include_dirs']=['/Developer/Panda3D/include/']

    elif lin:
        cythonBin=['cython']
        ext=".so"
        args['include_dirs']=['/usr/include/panda3d/']
    else:
        # Specify Panda3D include directory and
        # any other args here for other systems
        raise OSError("Specify compiler args for your system")

    args['language']="c++"
    
    build=c
    
    force=arg('force')
    fast=arg('fast')
    if fast: args['extra_compile_args'].append("-O0")

    if useOpenMP:
        args['extra_compile_args'].append('-fopenmp')
        args['extra_link_args'].append('-fopenmp')

    args['extra_compile_args'].extend(["-DNDEBUG"])
    args['extra_link_args'].extend(["libp3dtoolconfig.lib"])
    args['extra_link_args'].extend(["libp3dtool.lib"])
    args['extra_link_args'].extend(["libpandaexpress.lib"])
    if PROFILE:
        args['define_macros']=[('CYTHON_TRACE', '1')]

    cython_ext_modules=[]
    paths=[]
    compileDirs=['ncore','api','plugin','heapdict']
    if test == True:
        compileDirs.append('test')

    cythonExtensions=[".pyx"]
    cpy=arg('cpy')
    compileC=True#not cpy
    if cpy:
         cythonExtensions.append(".py")

    compileLoop(compileDirs, cythonBin, cythonExtensions, pyd_appends, ext, cython_ext_modules, paths, args)

def compileLoop(compileDirs, cythonBin, cythonExtensions, pyd_appends, ext, cython_ext_modules, paths, args):
    force=arg('force')
    build=os.curdir
    clean=arg('clean')
    compileC=True

    for root, dirs, files in itertools.chain.from_iterable(os.walk(d) for d in compileDirs):
        for f in files:
            import pdb
            pdb.set_trace()
            if any(f.endswith(e) for e in cythonExtensions):
                name='.'.join(f.split('.')[:-1])
                path=os.path.join(root,f)
                mod=os.stat(path).st_mtime
                dst=os.path.join(root,name+ext)
                src=os.path.join(build,name+pyd_appends+ext)

                cpath=os.path.join(root,name+'.cpp')
                source=path

                if clean:
                    remove(cpath)
                    remove(dst)
                    remove(os.path.join(root,name+'.pyc'))

                elif force or (not os.path.exists(dst)) or os.stat(dst).st_mtime<=mod:
                    remove(cpath)
                    
                    if f=="__init__.py":
                        print(("skipping "+src))
                        continue

                    cargs=cythonBin+["--cplus", source]
                    if verbose: print((' '.join(cargs)))
                    result=subprocess.call(cargs,shell=platformName=='Windows')
                    if result:
                        print(("\n\nERROR: Failed to cythonize "+src+"\n\n"))
                        return result
                    if not compileC: continue
                    print(("making Extension", dst))
                    if win:
                        extension=Extension(name, [path], **args)
                        extension.cython_directives={"language_level": 3}
                        #extension.cython_directives={'linetrace': True, 'binding':True}
                        #cython_ext_modules.append(extension)
                        #paths.append((src,dst))
                        error=compileModules([extension],[(src,dst)],name,build_ext)
                        if error: return error
                else:
                    if verbose: print((dst+" up to date"))
    
    return compileModules(cython_ext_modules,paths,name,build_ext)

def compileModules(cython_ext_modules,paths,name,build_ext):
    print('')
    setup(name = name,
        version = "1.0",
        ext_modules = cythonize(cython_ext_modules, compiler_directives={'linetrace': PROFILE, 'binding':PROFILE}),
        cmdclass = {'build_ext': build_ext}
    )

    print('')

    for src,dst in paths:
        print((src+" --> "+dst))
        shutil.move(src, dst)
    
    print('\n')
    
    return False

import collections
if __name__ == '__main__':

    if 'run' in sys.argv:
        run=True
        sys.argv.remove('run')
    else:
        run=False
    
    if 'build' in sys.argv:
        sys.argv.remove('build')
        t=time()
        err=protoBuild()
        if err:
            print(("Build Failed:",err))
            sys.exit(err)
        err=buildExtensions()
        print((time()-t))
        if err:
            print(("Build Failed:",err))
            sys.exit(err)
    if run:
        sys.argv=['main.py']
        import main
