import wx

import weakref

from panda3d.core import loadPrcFileData,WindowProperties,NodePath,ConfigVariableString
from direct.wxwidgets.WxPandaWindow import WxPandaWindow
from direct.showbase.ShowBase import ShowBase

loadPrcFileData("", "window-type none")
loadPrcFileData("", "framebuffer-hardware 0")
loadPrcFileData("", "sync-video 0")
loadPrcFileData("", "auto-flip 1")
loadPrcFileData('', 'basic-shaders-only 1')

import api.ops.ops

import time

extraRenderWorkaround=True # TODO : sort this out

def call2dict(**a): return a

def getBase():
    return PandaBase.instance if PandaBase.instance else PandaBase()

class PandaSplash():

    def __init__(self,node=None,winArgs={}):
        self.splashing=False
        self.splashed=False
        self.splashWin=None
        self.bestPipe=None
        self.bestScore=0.0
        self.winArgs=winArgs
        self.pipeSource=self.pipeGen()

    def scorePipe(self,pipe,win):
        # Check GFX Support:
        gsg=win.getGsg()
        if gsg.getSupportsBasicShaders():
            return 1.0 # good enough
        if gsg.isHardware():
            return 0.5 # kinda sucks
        return 0.1 # better than nothing

    def _close(self):
        if self.splashWin is not None:
            self.closeWindow(self.splashWin)
            self.splashWin=None

    def nextSplash(self):
        if self.bestScore>=1.0:
            return False # found one thats good enough
        for p in self.pipeSource:
            win=ShowBase.openWindow(self,pipe=p,**self.winArgs)
            if win is not None:
                #self.showImage(win)
                score=self.scorePipe(p,win)
                print(p.getInterfaceName(),"Scored: ",score)
                if score>self.bestScore:
                    self._close()
                    self.splashWin=win
                    self.bestScore=score
                    self.bestPipe=p
                    print("Selected: ",p.getInterfaceName())
                else:
                    self.closeWindow(win)
                return True
        # No more pipes!
        return False

    def splash(self):
        if self.splashing or self.splashed: return

        self.splashing=True

    def setBestPipe(self):
        """ set the best pipe. This may open lots of windows to find it """
        if not (self.splashing or self.splashed): self.splash()
        if not skipPipeSelect:
            if self.splashing:
                while self.nextSplash(): pass
            self.pipe=self.bestPipe
        print("Best Pipe Selected: ", self.pipe.getInterfaceName())
        #self.showImage()
    def endSplash(self):
        """
        calls setBestPipe then makes sure all splash windows are done and closed
        """
        self.setBestPipe()
        self.splashing=False
        self.splashed=False
        self._close()

    def pipeGen(self):
        """
        Creates all GraphicsPipes that the system knows about and fill up
        self.pipeList with them.
        """
        from panda3d.core import GraphicsPipeSelection
        selection = GraphicsPipeSelection.getGlobalPtr()

        for name in pipeNames:
            pipe=selection.makeModulePipe(name)
            if pipe:
                self.notify.info("Got graphics pipe %s (%s)." % (
                    pipe.getType().getName(), pipe.getInterfaceName()))
                self.pipeList.append(pipe)
                yield pipe
            else:
                self.notify.warning("Could not make graphics pipe %s." % (name))

    def showImage(self,win):
        return
        if win is None: return
        #image="Images/SplashScreen.png"
        #camera2d=self.makeCamera2d(win)
        #render2d=NodePath("render2d")
        #camera2d.node().setScene(render2d)
        #self.myImage=OnscreenImage(image=image, pos = (0, 0, 0),parent=render2d, scale=(1,1,1))
        #for i in range(1): 
        #    self.graphicsEngine.renderFrame()

        #self.myImage.destroy()


class PandaBase(ShowBase):
    instance=None

    def __init__(self):
        title="Loading {0}...".format("fudged_import")
        props=WindowProperties.getDefault()
        props.setTitle(title)
        props.setForeground(False)
        d=call2dict(props = props, fbprops = None,
                   type = None, name = title, size = (750,375),
                   scene = None)
        #PandaSplash.__init__(self,winArgs=d)
        self.checkedGFX=False

        self.props=props

        self.leakDict=weakref.WeakKeyDictionary()
        #print('-- instancing base--')
        
        #if self.__class__.instance:
        #    print("Error: Making more than one PandaBase")
        self.__class__.instance=self

        ShowBase.__init__(self)

        self.wxApp=None
        self.frame=None

        # the last cadwin to get an activate event
        #self.lastActive=None
        # used for focus detection on mac
        #self.lastActiveView3d=None

        self.taskMgr.removeTasksMatching('igLoop')
        
        self.frameSort=-100
        self.taskMgr.add(self.doFrame,'doFrames',sort=self.frameSort)

        #self.accept("f", self.fps)
        self.frameTime=time.time()
        self.fpsTime=time.time()
        self.frameCount=0
        
        self.cadWins=[]
        self.viewWins=[]
        self.view3ds=[]

        ##app=api.startup_bar.getAppShell()
        ##startup_bar.stop(True)
        ##app.wxStep()
        ##self.setBestPipe()

        #self.direct = None

        if extraRenderWorkaround:
            self.framesSinceExtraRender=0
        #self.disableMouse()

    def updateProps(self):
        self.win.requestProperties(self.props)

    def startWx(self, *args, **kwargs):
        # override base.startWx, since we don't need it getting called
        # maybe this is useful.
        print("START WX")
        #pass

    def userExit(self):
        print("Exiting. Shutdown...")
        ShowBase.userExit(self)
        sys.exit()

    def getFrontCadWin(self):
        # this is kinda hacky and will fail sometimes
        
        if len(self.cadWins)==1:
            return self.cadWins[0]
        elif len(self.cadWins)==0:
            return None
        else:
            if self.lastActive in self.cadWins:
                return self.lastActive
            else:
                print("Error: PandaBase.getFrontCadWin failed")
                return None
    
    def checkGFX(self,win):
        # Check GFX Support:
        gsg=win.getGsg()
        def err(txt):
            from api import messagebox
            gfxErrorText=" Attempting to continue, but this error may cause graphical artifacts, crashes or other problems."
            messagebox.error(txt+gfxErrorText)
        if not gsg.getSupportsBasicShaders():
            if not skipShaderSupportWarning:
                err("Video driver reports that shaders are not supported.")
        if not gsg.isHardware():
            err("Video driver appears to be using a software renderer. For optimal performance, a hardware accelerated renderer is recommended.")

    def fps(self):
        print(self.frameCount/(time.time()-self.fpsTime))
        self.fpsTime=time.time()
        self.frameCount=0

    def setWxApp(self, wxApp):
        self.wxApp=wxApp

    def setFrame(self):
        if self.wxApp is not None:
            self.frame=self.wxApp.GetTopWindow()
            self.attrNames=["HasFocus", "HasCapture", "IsActive", "FindFocus"]
            self.attrVals=[None, None, None, None]
            frame=self.frame
            self.attrMethods=[frame.HasFocus, frame.HasCapture, frame.IsActive, frame.FindFocus]
            for n in self.attrNames:
                print str(n) + '\t',
            print

    def printWinAttrs(self):
        if self.frame is not None:
            show=False
            for i, (n, v, m) in enumerate(zip(self.attrNames, self.attrVals, self.attrMethods)):
                vnew=m()
                if v != vnew:
                    show=True
                    self.attrVals[i]=vnew
            if show:
                for v in self.attrVals:
                    print str(v) + '\t',
                print
    
    def doFrame(self,task):
        self.printWinAttrs()
        #self.updateProps()
        if len(self.leakDict)>0:
            print("Memory Leak:", [(sys.getrefcount(k),v) for k, v in self.leakDict.items()])
            try:
                import objgraph
                print("Writing leak graph...")
                objgraph.show_backrefs(list(self.leakDict.keys()), filename='leak-backref-graph.svg', max_depth=4, too_many=40, refcounts=True)
            except:
                print("Could not write leak graph")
            self.leakDict.clear()

        # TODO : figure out what is running the wx events on windows.
        # On windows, most of the wx events (all but menu shortcuts it seems) are run my something else
        self.wxApp.wxStep()
        numActive=0

        for c in self.view3ds:
            if not self.checkedGFX:
                self.checkGFX(c.win)
                self.checkedGFX=True
            c._checkEvents()
        if len(self.cadWins)==0 and len(self.viewWins)==0:# and not allowNoWindowsOpen:
            self.userExit()
        for c in self.view3ds:
            c._preRender()
            if c.win.isActive():
                numActive+=1
        
        if extraRenderWorkaround:
            countBeforeRender=16
            if numActive==0:
                if self.framesSinceExtraRender>countBeforeRender:
                    for c in self.view3ds:
                        c.win.setActive(True)
                    self.framesSinceExtraRender=0
            else:
                self.framesSinceExtraRender=countBeforeRender-1
            self.framesSinceExtraRender+=1

        self.graphicsEngine.renderFrame()

        minFrameLength=.02 if numActive==0 else .005
        t=time.time()
        s=minFrameLength-(t-self.frameTime)
        if s>0:
            time.sleep(s)

        self.frameCount+=1
        self.frameTime=time.time()
            
        return task.cont

    def setActive(self,cadWin):
        #print "ncore.pandabase.PandaBase:setActive: A"
        if cadWin in self.cadWins:
            self.lastActive=cadWin

    def removeCadWin(self,cadWin):
        if self.lastActive==cadWin: self.lastActive=None
        for i,w in enumerate(self.cadWins):
            if w==cadWin:
                del self.cadWins[i]
                return
        print("Error: failed to close window")

    def removeView3d(self,view):
        if self.lastActiveView3d==view:
            self.lastActiveView3d=None
        for i,w in enumerate(self.view3ds):
            if w==view:
                del self.view3ds[i]
                return
        print("Error: failed to close view3d")

    def removeViewWin(self,viewWin):
        for i,w in enumerate(self.viewWins):
            if w==viewWin:
                del self.viewWins[i]
                return
        print("Error: failed to close viewWin")

class PandaPanel(WxPandaWindow):
    
    def __init__( self, *args, **kargs ):
        WxPandaWindow.__init__(self, *args, **kargs)
        self.Bind( wx.EVT_PAINT, self.pandaPaint )
        self.view3d=None

    def setView(self,view3d):
        self.view3d=view3d

    def pandaPaint(self,event):
        if self.view3d is not None: 
            self.view3d.setNeedRender()
        event.Skip()

def getAppShell():
    return AppShell.instance if AppShell.instance else AppShell()

class AppShell(wx.PySimpleApp):
    instance=None
    def __init__(self):
        if self.__class__.instance:
            error("Making more than one AppShell")
        self.__class__.instance=self
        
        wx.PySimpleApp.__init__(self)#,redirect=True)
        self.SetAppName("Minimal Example")
        self.appInit()
        getBase().setWxApp(self)
        
    def queryEnd(self, event):
        print("queryEnd")
        if not askQuit():
            event.Veto(True)
        print("queryEnded")
    
    def wxStep(self, task = None):
        """A step in the WX event loop. You can either call this yourself or use as task."""
        while self.evtLoop.Pending():
            self.evtLoop.Dispatch()
        self.ProcessPendingEvents()
        if task != None: return task.cont

    def appInit(self):
        # Create a new event loop (to overide default wxEventLoop)
        self.evtLoop = wx.EventLoop()
        self.oldLoop = wx.EventLoop.GetActive()
        wx.EventLoop.SetActive(self.evtLoop)
        
        # TODO : fix binding these does not seem to do anything!
        # should catch quit from dock
        #wx.GetApp().Bind(wx.EVT_QUERY_END_SESSION, self.queryEnd)
        ##wx.GetApp().Bind(wx.EVT_END_SESSION, self.queryEnd)
